//###########################################################
// File: mmc_spi.c
//
// Readroutines for MMC MultiMedia cards and
// SD SecureDigital cards in SPI mode.
//
// This will work only for MMC cards with 512 bytes block length !
// This will work only for MMC cards with a partition table !
//
// 10.10.2006 Added a busy check after CMD0. Without this one of
//            my newer 512MB NoName card hangs.
//
// 25.09.2006 Split SPI_WRITE() into SPI_WRITE() and SPI_WAIT() for
//            better compatibility with my other mmc_spi.c versions ;)
//
// 04.09.2006 Made new MMC_IO_Init(). Removed IO pin settings from
//            MMCIdentify(). Why ? See comments above MMC_IO_Init(). 
//
// 22.08.2005 Bugfix for SanDisk 256MB SD card in MMCIdentify();
//
//
// Benutzung auf eigene Gefahr !
//
// Use at your own risk !
//
//#########################################################################
// Last change: 04.09.2006
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 3.4.5
//#########################################################################

#include <avr/io.h>

#include "dos.h"

#ifdef MMC_CARD_SPI

//######################################################
unsigned char MMCCommand(unsigned char command, unsigned long adress)
//######################################################
{

 SPI_WRITE(DUMMY_WRITE); 			 //Dummy write
 SPI_WAIT();
 SPI_WRITE(command);
 SPI_WAIT();
 SPI_WRITE((unsigned char)(adress>>24)); //MSB of adress
 SPI_WAIT();
 SPI_WRITE((unsigned char)(adress>>16));
 SPI_WAIT();
 SPI_WRITE((unsigned char)(adress>>8));
 SPI_WAIT();
 SPI_WRITE((unsigned char)adress);       //LSB of adress
 SPI_WAIT();
 SPI_WRITE(0x95); 			 // Checksum for CMD0 GO_IDLE_STATE and dummy checksum for other commands
 SPI_WAIT();

 //wait for response
 while(SPI_REGISTER==DUMMY_WRITE)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
  }

 return SPI_REGISTER;
}

//######################################################
unsigned char MMCStartReadSector(unsigned long sector)
//######################################################
{
 unsigned char by;
 unsigned long startadr;
 
 if(sector>=maxsect) return 1; //sectornumber too big

//calculate startadress of the sector
 startadr=sector * BYTE_PER_SEC;

 by=MMCCommand(MMC_READ_BLOCK,startadr);

 while(by!=START_BLOCK_TOKEN)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_REGISTER; // wait for card response
  }
  
  return 0;
}

//######################################################
// Init MMC IO pins
//
// Removed this from MMCIdentify().
// Maybe some other devices are connected to SPI bus.
// All chip select pins of these devices should have
// high level before starting SPI transmissions !
// So first call MMC_IO_Init(), after that for example
// VS1001_IO_Init(), SPILCD_IO_Init() and AFTER that MMCIdentify().
void MMC_IO_Init(void)
//######################################################
{
 sbi(MMC_CS_PORT,MMC_CS_BIT); //MMC_CS set 1
 sbi(MMC_CS_DDR,MMC_CS_BIT);  //MMC_CS output

 cbi(MMC_SCK_PORT,MMC_SCK_BIT); //SCK set 0
 sbi(MMC_SCK_DDR,MMC_SCK_BIT);  //SCK output

 cbi(MMC_MISO_DDR,MMC_MISO_BIT);  //MISO input

 cbi(MMC_MOSI_PORT,MMC_MOSI_BIT); //MOSI set 0
 sbi(MMC_MOSI_DDR,MMC_MOSI_BIT);  //PB5 MOSI output
}

//######################################################
unsigned char MMCIdentify(void)
//######################################################
{
 unsigned char by;
 unsigned int i;
 unsigned int c_size, c_size_mult, read_bl_len;
 unsigned long drive_size;

//Init SPI with a very slow transfer rate first !

//SPCR SPI Controlregister
// SPIE=0; //No SPI Interrupt
// SPE=1;  //SPI Enable
// DORD=0; //Send MSB first
// MSTR=1; //I am the master !
// CPOL=0; //SCK low if IDLE
// CPHA=0; //SPI Mode 0
// SPR1=1; //SPI Clock = f/128 = 125kHz @16MHz Clock
// SPR0=1; //or f/64 if SPI2X = 1 in SPSR register
 SPCR=0x53;

//SPSR SPI Statusregister
// SPI2X=1; //Double speed for SPI = 250kHz @16MHz Clock
// SPSR=0x01;
 SPSR=0x00;

 for(i=0; i<10; i++) { SPI_WRITE(DUMMY_WRITE);  SPI_WAIT(); } // give min 74 clock pulses before
                                      // sending commands
 
 MMC_CS_OFF();

//send CMD0 for RESET
 by=MMCCommand(MMC_GO_IDLE_STATE,0);
 while(by==0x00) //wait til busy is gone
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_REGISTER;
  } 

//repeat CMD1 til result=0
 do
  {
   by=MMCCommand(MMC_INIT,0);
  }while(by!=0);

//read CID
// MMCCommand(MMC_READ_CID,0); // nothing really interesting here

//read CSD Card Specific Data
 by=MMCCommand(MMC_READ_CSD,0);

// This waiting loop is for SanDisk 256MB SD
// Works with my other cards too
 while(by!=START_BLOCK_TOKEN)
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_REGISTER; // wait for card response
  }

 for(i=0; i<16; i++) //CSD has 128 bits -> 16 bytes
  {
   SPI_WRITE(DUMMY_WRITE);
   SPI_WAIT();
   by=SPI_REGISTER;
   dirbuf[i]=by;
  }

 SPI_WRITE(DUMMY_WRITE); // 16 bit crc follows data
 SPI_WAIT();
 SPI_WRITE(DUMMY_WRITE);
 SPI_WAIT();

 c_size=dirbuf[6] & 0x03; //bits 1..0
 c_size<<=10;
 c_size+=(unsigned int)dirbuf[7]<<2;
 c_size+=dirbuf[8]>>6;

 by= dirbuf[5] & 0x0F;
 read_bl_len=1;
 read_bl_len<<=by;

 by=dirbuf[9] & 0x03;
 by<<=1;
 by+=dirbuf[10] >> 7;
 
 c_size_mult=1;
 c_size_mult<<=(2+by);

 drive_size=(unsigned long)(c_size+1) * (unsigned long)c_size_mult * (unsigned long)read_bl_len;
 maxsect= drive_size / BYTE_PER_SEC;

 MMC_CS_ON();

//switch to high speed SPI
// SPR1=0; //SPI Clock = f/4 = 4MHz @16MHz Clock
// SPR0=0; //or f/2 if SPI2X = 1 in SPSR register
 SPCR=0x50;

//SPSR SPI Statusregister
// SPI2X=1; //Double speed for SPI = 8MHz @16MHz Clock
 SPSR=0x01;

 return 0;
}
#endif //MMC_CARD_SPI
